@@ -1,13 +1,12 @@
 package Apache::SSI;
 
 use strict;
-use vars qw($VERSION @ISA);
-use HTML::SimpleParse;
+use vars qw($VERSION);
 use Apache::Constants qw(:common OPT_EXECCGI);
 use File::Basename;
+use HTML::SimpleParse;
 
-$VERSION = '1.94';
-@ISA = qw(HTML::SimpleParse);
+$VERSION = '1.95';
 my $debug = 0;
 
 sub handler($$) {
@@ -26,7 +25,7 @@ sub handler($$) {
 
 	unless (-e $file) {
 		$r->log_error("$file not found");
-		return NOT_FOUND;	
+		return NOT_FOUND;
 	}
 
 	local *IN;
@@ -41,34 +40,62 @@ sub handler($$) {
 }
 
 sub new {
-	my $self = $_[0]->SUPER::new($_[1]);
-	$self->{_r} = $_[2];
-	return $self;
+	my ($pack, $text, $r) = @_;
+	return bless {
+		'text' => $text,
+		'_r'   => $r,
+	}, $pack;
 }
 
-sub lastmod {
+sub get_output($) {
+	my $self = shift;
+	
+	my $out = '';
+	my @parts = split m/(<!--#.*?-->)/s, $self->{'text'};
+	while (@parts) {
+		$out .= shift @parts;
+		last unless @parts;
+		my $ssi = shift @parts;
+		if ($ssi =~ m/^<(!--#.*--)>$/) {
+			$out .= $self->output_ssi($1);
+		} else { die 'Parse error' }
+	}
+	return $out;
+}
+
+
+sub output($) {
+	my $self = shift;
+	
+	my @parts = split m/(<!--#.*?-->)/s, $self->{'text'};
+	while (@parts) {
+		print shift @parts;
+		last unless @parts;
+		my $ssi = shift @parts;
+		if ($ssi =~ m/^<(!--#.*--)>$/) {
+			print $self->output_ssi($1);
+		} else { die 'Parse error' }
+	}
+}
+
+sub lastmod($) {
 	return scalar localtime( (stat $_[0])[9] );
 }
 
-
-# Method called by HTML::SimpleParse::output, overrides
-# HTML::SimpleParse::output_ssi.
-sub output_ssi {
-	my $self = shift;
-	my $text = shift;
+sub output_ssi($$) {
+	my ($self, $text) = @_;
 	
 	if ($text =~ s/^!--#(\w+)\s*//) {
 		my $method = lc "ssi_$1";
 		$text =~ s/--$//;
-		no strict('refs');
 		warn "returning \$self->$method(...)" if $debug;
-		my $args = [ $self->parse_args($text) ];
+		my $args = [ HTML::SimpleParse->parse_args($text) ];
 		return $self->$method( {@$args}, $args );
 	}
 	return;
 }
 
-sub ssi_include {
+sub ssi_include($$) {
 	my ($self, $args) = @_;
 	my $r = $self->{_r};
 
@@ -79,21 +106,29 @@ sub ssi_include {
 	return;
 }
 
-sub ssi_fsize { 
+sub ssi_fsize($$) { 
 	my ($self, $args) = @_;
 	return -s $self->find_file(@{$args}{'file', 'virtual'});  # $ for BBEdit
 }
 
-sub ssi_flastmod {
+sub ssi_flastmod($$) {
 	my($self, $args) = @_;
-	return &lastmod($args->{file} || $self->{_r}->filename);
+	my $req;
+	if (exists $args->{file}) {
+		$req = $self->{_r}->lookup_file( $args->{file} );
+	} elsif (exists $args->{virtual}) {
+		$req = $self->{_r}->lookup_uri($args->{virtual});
+	} else {
+		$req = $self->{_r};
+	}
+	return &lastmod( $req->filename() );
 }
 
-sub ssi_printenv {
+sub ssi_printenv() {
 	return join "", map( {"$_: $ENV{$_}<br>\n"} keys %ENV );
 }
 
-sub ssi_exec {
+sub ssi_exec($$) {
 	my($self, $args) = @_;
 	#XXX did we check enough?
 	my $r = $self->{_r};
@@ -105,62 +140,56 @@ sub ssi_exec {
 	return scalar `$args->{cmd}`;
 }
 
-sub ssi_perl {
+sub ssi_perl($$$) {
 	my($self, $args, $margs) = @_;
-	local $_;
-	my (@arg1, @arg2);
+
+	my ($pass_r, @arg1, @arg2, $sub) = (1);
 	{
 		my @a;
 		while (@a = splice(@$margs, 0, 2)) {
 			if ($a[0] eq 'sub') {
-				$_ = $a[1];
+				$sub = $a[1];
 			} elsif ($a[0] eq 'arg') {
 				push @arg1, $a[1];
 			} elsif ($a[0] eq 'args') {
 				push @arg1, split(/,/, $a[1]);
+			} elsif (lc $a[0] eq 'pass_request') {
+				$pass_r = 0 if lc $a[1] eq 'no';
 			} else {
 				push @arg2, @a;
 			}
 		}
 	}
 
-	my $sub;
-	if ( /^\s*sub[^\w:]/ ) {     # for <!--#perl sub="sub {print ++$Access::Cnt }" -->
-		$sub = eval();
+	if ( $sub =~ /^\s*sub[^\w:]/ ) {     # for <!--#perl sub="sub {print ++$Access::Cnt }" -->
+		$sub = eval($sub);
 	} else {             # for <!--#perl sub="package::subr" -->
-		$sub = (/::/ ? $_ : "main::$_");
+		$sub = "main::$_" unless $sub =~ /::/;
 	}
+	
+	$pass_r = 0 if $self->{_r} and lc $self->{_r}->dir_config('SSIPerlPass_Request') eq 'no';
+	unshift @arg1, $self->{_r} if $pass_r;
 	warn "sub is $sub, args are @arg1 & @arg2" if $debug;
 	no strict('refs');
 	return scalar &{ $sub }(@arg1, @arg2);
 }
 
-sub multi_args {
-	shift;  # Get rid of $self
-	my $arg = shift; # What arg to look for
-	my @list = @{shift()};
-	my (@returns, @pair);
-	while (@pair = splice(@list, 0, 2)) {
-		push(@returns, $pair[1]) if $pair[0] eq $arg;
-	}
-	return @returns;
-}
-
-sub ssi_set {
+sub ssi_set($$) {
 	my ($self, $args) = @_;
 	
 	# Work around a bug in mod_perl 1.12 that happens when calling
 	# subprocess_env in a void context
-	my $trash = $self->{_r}->subprocess_env( $args->{var}, $args->{value} );
+	#my $trash = $self->{_r}->subprocess_env( $args->{var}, $args->{value} );
+	$self->{_r}->subprocess_env( $args->{var}, $args->{value} );
 	return;
 }
 
-sub ssi_config {
+sub ssi_config() {
 	warn "*** 'config' directive not implemented by Apache::SSI";
 	return "<$_[1]>";
 }
 
-sub ssi_echo {
+sub ssi_echo($$) {
 	my($self, $args) = @_;
 	my $var = $args->{var};
 	my $value;
@@ -176,11 +205,11 @@ sub ssi_echo {
 	return '';
 }
 
-sub echo_DATE_GMT { scalar gmtime; }
-sub echo_DATE_LOCAL { scalar localtime; }
-sub echo_DOCUMENT_NAME { basename $_[0]->filename; }
-sub echo_DOCUMENT_URI { $_[0]->uri; }
-sub echo_LAST_MODIFIED { &lastmod($_[0]->filename); }
+sub echo_DATE_GMT() { scalar gmtime; }
+sub echo_DATE_LOCAL() { scalar localtime; }
+sub echo_DOCUMENT_NAME($) { basename $_[0]->filename; }
+sub echo_DOCUMENT_URI($) { $_[0]->uri; }
+sub echo_LAST_MODIFIED($) { &lastmod($_[0]->filename); }
 
 1;
 
@@ -286,7 +315,21 @@ The "args" parameter will simply split on commas, meaning that currently
 there's no way to embed a comma in arguments passed via the "args"
 parameter.  Use the "arg" parameter for this.
 
-See C<http://perl.apache.org/src/mod_perl.html> for more details on this.
+If you give a key-value pair and the key is not 'sub', 'arg', 'args', or 
+'pass_request' (see below), then your routine will be passed B<both> the 
+key and the value.  This lets you pass a hash of key-value pairs to your 
+function:
+
+ <!--#perl sub=holy::matrimony groom=Hi bride=Lois-->
+ Will call &holy::matrimony('groom', 'Hi', 'bride', 'Lois');
+
+As of version 1.95, we pass the current Apache request object ($r) as the
+first argument to the function.  To turn off this behavior, give the key-value
+pair 'pass_request=no', or put 'PerlSetVar SSIPerlPass_Request no' in your
+server's config file.
+
+See C<http://perl.apache.org/src/mod_perl.html> for more information on Perl
+SSI calls.
 
 =item * config
 
@@ -311,15 +354,15 @@ variables mod_include supplies.  Is this the correct order?
 
 =head1 TO DO
 
+Make sure things that use virtual paths are doing them correctly.  (Ibid)
+
+Revisit http://www.apache.org/docs/mod/mod_include.html and see what else
+there I can implement.
+
 It would be nice to have a "PerlSetVar ASSI_Subrequests 0|1" option that
 would let you choose between executing a full-blown subrequest when
 including a file, or just opening it and printing it.
 
-It would also be nice to mix & match the "arg" and "args" parameters to
-<!--#perl--> sections, like so:
-
- <!--#perl sub=something arg="Hi, Ken" args=5,12,13 arg="Bye, Ken"-->
-
 I'd like to know how to use Apache::test for the real.t test.
 
 =head1 BUGS
@@ -336,7 +379,8 @@ Apache::OutputChain(3)
 
 Ken Williams ken@forum.swarthmore.edu
 
-Based on original version by Doug MacEachern dougm@osf.org
+Concept based on original version by Doug MacEachern dougm@osf.org, 
+implementation different.
 
 =head1 COPYRIGHT
 
